/*
 * base64.c -- base-64 conversion routines.
 *
 * For license terms, see the file COPYRIGHT in the parent directory.
 *
 * This base 64 encoding is defined in RFC2045 section 6.8,
 * "Base64 Content-Transfer-Encoding", but lines must not be broken in the
 * scheme used here.
 */

#ifndef BASE64_H
#define BASE64_H

void to64frombits(unsigned char *out, const unsigned char *in, int inlen);
int from64tobits(char *out, const char *in, int maxlen);

#endif